from typing import Iterable, List, Optional

import torch
from transformers import BertTokenizerFast

from rex.data.collate_fn import GeneralCollateFn
from rex.data.label_encoder import LabelEncoder
from rex.data.transforms.base import TransformBase
from rex.data.vocab import Vocab, get_pad_mask, get_pad_token
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar


class CachedTaggingTransform(TransformBase):
    """Cached data transform for classification task."""

    def __init__(self, max_seq_len: int, label2id_filepath: str, plm_dir: str) -> None:
        super().__init__()

        self.max_seq_len = max_seq_len
        self.label_encoder = LabelEncoder.from_pretrained(label2id_filepath)
        self.tokenizer = BertTokenizerFast.from_pretrained(plm_dir)

        self.collate_fn = GeneralCollateFn(
            {
                # None means keep its original type
                "id": None,
                "tokens": None,
                "ner_tags": None,
                "token_ids": torch.long,
                "mask": torch.long,
                "labels": torch.long,
            }
        )

    def transform(
        self,
        dataset: Iterable,
        desc: Optional[str] = "Transform",
        debug: Optional[bool] = False,
        **kwargs,
    ) -> List[dict]:
        final_data = []
        if debug:
            dataset = dataset[:500]
        transform_loader = pbar(dataset, desc=desc)

        num_tot_ins = 0
        for data in transform_loader:
            tokens = data["tokens"]
            token_ids = self.tokenizer.convert_tokens_to_ids(tokens)
            token_ids = get_pad_token(
                token_ids, self.max_seq_len, self.tokenizer.pad_token_id
            )
            mask = get_pad_mask(len(token_ids), self.max_seq_len)
            labels = data["ner_tags"]
            label_ids = self.label_encoder.encode(labels, update=False)
            label_ids = get_pad_token(
                label_ids, self.max_seq_len, self.label_encoder.label2id["O"]
            )

            assert len(token_ids) == self.max_seq_len
            assert len(mask) == self.max_seq_len
            assert len(label_ids) == self.max_seq_len

            ins = {
                "id": data["id"],
                "tokens": data["tokens"],
                "ner_tags": data["ner_tags"],
                "token_ids": token_ids,
                "mask": mask,
                "labels": label_ids,
            }
            final_data.append(ins)

        logger.info(transform_loader)
        logger.info(f"#Total Ins: {num_tot_ins}")

        return final_data

    def predict_transform(self, text: str):
        """
        Args:
            data:
                {
                    "id": "blahblah",
                    "text": "text",
                }
        """
        token_ids = self.tokenizer.convert_tokens_to_ids(list(text))
        token_ids = get_pad_token(token_ids, self.max_seq_len, self.tokenizer.pad)
        mask = get_pad_mask(len(token_ids), self.max_seq_len)

        obj = {
            "tokens": list(text),
            "token_ids": token_ids,
            "mask": mask,
        }
        return obj



"""
Scientific number
"""

import math

import numpy as np


def is_close(num1, num2, rel_tol=None, abs_tol=None):
    return math.isclose(num1, num2, rel_tol=rel_tol, abs_tol=abs_tol)


def convert_numpy_obj_to_native(obj):
    if isinstance(obj, dict):
        for key in obj:
            convert_numpy_obj_to_native(obj[key])



from typing import Callable, List

from rex.data.vocab import Vocab
from rex.utils.config import ConfigArgument, ConfigParser
from rex.utils.io import load_embedding_file, load_line_iterator
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar


def build_vocab(tokenize_func: Callable, lines: List[str]):
    vocab = set()
    for line in lines:
        tokens: List[str] = tokenize_func(line.strip())
        vocab.update(tokens)
    return vocab


def build_emb(
    raw_emb_filepath: str,
    dump_emb_filepath: str,
    *files: str,
    tokenize_func: Callable = None,
):
    """build vocab and filter out useless embeddings from ``raw_emb_filepath``"""
    logger.info(f"loading embedding from {raw_emb_filepath}")
    tokens, token2vec = load_embedding_file(raw_emb_filepath)
    logger.info(f"#tokens in emb file: {len(tokens)}")

    if tokenize_func is None:
        logger.info("Use char tokenize function.")
        tokenize_func = list

    logger.info("building vocab")
    raw_vocab = set()
    for filepath in pbar(files):
        for line in load_line_iterator(filepath):
            ins_tokens = tokenize_func(line.strip())
            raw_vocab.update(ins_tokens)
    logger.info(f"vocab size of corpora: {len(raw_vocab)}")

    logger.info("updating vocab")
    final_tokens = raw_vocab & set(tokens)
    logger.info(f"final vocab size of the intersection: {len(final_tokens)}")
    final_vocab = Vocab(init_pad_unk_emb=True)
    for token in pbar(final_tokens):
        final_vocab.add(token, token2vec[token])

    logger.info("dumping vocab")
    final_vocab.save_pretrained(dump_emb_filepath, dump_weights=True)


if __name__ == "__main__":
    config = ConfigParser.parse_cmd(
        ConfigArgument("-r", "--raw-emb-filepath", help="raw embedding filepath"),
        ConfigArgument("-o", "--dump-emb-filepath", help="dumped embedding filepath"),
        ConfigArgument("-f", "--filepaths", nargs="*", help="raw text filepaths"),
    )
    build_emb(config.raw_emb_filepath, config.dump_emb_filepath, *config.filepaths)



from typing import Optional

import torch
import torch.nn as nn

from rex.utils.io import load_embedding_file


class StaticEmbedding(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        dropout: Optional[float] = 0.3,
    ) -> None:
        super().__init__()

        self.dropout = nn.Dropout(dropout)
        self.embedding = nn.Embedding(num_embeddings, embedding_dim)

    def forward(self, token_ids):
        rep = self.embedding(token_ids)
        rep = self.dropout(rep)
        return rep

    @classmethod
    def from_pretrained(
        cls,
        filepath: str,
        file_encoding: Optional[str] = "utf8",
        freeze: Optional[bool] = False,
    ):
        tokens, token2emb = load_embedding_file(filepath, encoding=file_encoding)
        weights = []
        for token in tokens:
            weights.append(token2emb[token])
        weights = torch.tensor(weights, dtype=torch.float)
        embed = cls(weights.shape[0], weights.shape[1])
        embed.embedding = nn.Embedding.from_pretrained(weights, freeze=freeze)
        return embed



__all__ = ["__version__", "data", "metrics", "models", "modules", "tasks", "utils"]

from accelerate import Accelerator, DistributedDataParallelKwargs

from rex.version import __version__

ddp_args = DistributedDataParallelKwargs(find_unused_parameters=True)
accelerator = Accelerator(kwargs_handlers=[ddp_args])



import torch.nn as nn
import torch.nn.functional as F
from transformers import BertModel

from rex.modules.crf import PlainCRF


class LSTMCRFModel(nn.Module):
    def __init__(self, plm_dir, num_lstm_layers, num_tags, dropout):
        super().__init__()

        self.num_tags = num_tags

        self.plm = BertModel.from_pretrained(plm_dir)
        self.hidden_size = self.plm.config.hidden_size
        # self.lstm_enc = nn.LSTM(
        #     self.hidden_size,
        #     self.hidden_size // 2,
        #     num_layers=num_lstm_layers,
        #     bias=True,
        #     batch_first=True,
        #     dropout=dropout,
        #     bidirectional=True,
        # )
        self.hidden2tag = nn.Linear(self.hidden_size, num_tags)
        self.dropout = nn.Dropout(dropout)
        # self.crf = PlainCRF(num_tags)

    def forward(self, token_ids, mask=None, labels=None, **kwargs):
        out = self.plm(input_ids=token_ids, attention_mask=mask, return_dict=True)
        out = out.last_hidden_state
        # out, (_, _) = self.lstm_enc(emb)
        # out = self.dropout(out)
        out = self.hidden2tag(out)

        # results = {"pred": self.crf.decode(out)}
        results = {"pred": out.argmax(-1).detach().cpu().tolist()}
        if labels is not None:
            labels[mask == 0] = -1
            results["loss"] = F.cross_entropy(
                out.reshape(-1, self.num_tags), labels.reshape(-1), ignore_index=-1
            )
            # -self.crf(out, labels, mask, reduction="mean")
            # results["loss"] = -self.crf(out, labels, mask, reduction="mean")

        return results



from accelerate.state import AcceleratorState
from loguru._logger import Core as _Core
from loguru._logger import Logger as _Logger
from tqdm import tqdm as _tqdm


class MultiProcessLogger(_Logger):
    def __init__(self, *args, main_process_logging=True):
        super().__init__(*args)

        self.main_process_logging = main_process_logging

    @staticmethod
    def _should_log(main_process_only):
        "Check if log should be performed"
        return not main_process_only or (
            main_process_only and AcceleratorState().local_process_index == 0
        )

    def _log(self, *args):
        """
        Delegates logger call after checking if we should log.

        Accepts a new kwarg of `main_process_only`, which will dictate whether it will be logged across all processes
        or only the main executed one. Default is `True` if not passed
        """
        if self._should_log(self.main_process_logging):
            super()._log(*args)


# logger depth=1 to show original filename and call info under logger inheritance
logger = MultiProcessLogger(_Core(), None, 1, False, False, True, False, True, None, {})
logger.add(
    lambda msg: _tqdm.write(msg, end=""), colorize=True, backtrace=True, diagnose=True
)



import torch.optim as optim

from rex.data.data_manager import DataManager
from rex.data.dataset import CachedDataset
from rex.metrics.tagging import tagging_prf1
from rex.tasks.simple_task import SimpleTask
from rex.utils.io import load_jsonlines
from rex.utils.registry import register
from rex.utils.tagging import get_entities_from_tag_seq

from .model import LSTMCRFModel
from .transform import CachedTaggingTransform


@register("task")
class TaggingTask(SimpleTask):
    def __init__(self, config) -> None:
        super().__init__(config)

    def init_transform(self):
        return CachedTaggingTransform(
            self.config.max_seq_len,
            self.config.label2id_filepath,
            self.config.plm_dir,
        )

    def init_data_manager(self):
        return DataManager(
            self.config.train_filepath,
            self.config.dev_filepath,
            self.config.test_filepath,
            CachedDataset,
            self.transform,
            load_jsonlines,
            self.config.train_batch_size,
            self.config.eval_batch_size,
            self.transform.collate_fn,
            use_stream_transform=False,
            debug_mode=self.config.debug_mode,
            dump_cache_dir=self.config.dump_cache_dir,
        )

    def init_model(self):
        m = LSTMCRFModel(
            self.config.plm_dir,
            num_lstm_layers=self.config.num_lstm_layers,
            num_tags=self.transform.label_encoder.num_tags,
            dropout=self.config.dropout,
        )
        return m

    def init_optimizer(self):
        return optim.Adam(self.model.parameters(), lr=self.config.learning_rate)

    def get_eval_results(self, input_batches: list, output_batches: list) -> dict:
        preds = []
        golds = []
        for batch_input, batch_output in zip(input_batches, output_batches):
            batch_tokens = batch_input["tokens"]
            batch_gold = batch_input["ner_tags"]
            batch_pred = batch_output["pred"]
            for tokens, ins_gold, ins_pred in zip(batch_tokens, batch_gold, batch_pred):
                ins_gold = ins_gold[: len(tokens)]
                ins_pred = ins_pred[: len(tokens)]
                ins_gold_ents = get_entities_from_tag_seq(tokens, ins_gold)
                ins_pred = self.transform.label_encoder.decode(ins_pred)
                ins_pred_ents = get_entities_from_tag_seq(tokens, ins_pred)
                golds.append(ins_gold_ents)
                preds.append(ins_pred_ents)
        measures = tagging_prf1(golds, preds)
        return measures

    def predict_api(self, *texts):
        raw_dataset = [self.transform.predict_transform(text) for text in texts]
        loader = self.data_manager.prepare_loader(raw_dataset)
        results = []
        for batch in loader:
            batch_out = self.model(**batch)
            for tokens, ins_pred in zip(batch["tokens"], batch_out["pred"]):
                ins_pred = ins_pred[: len(tokens)]
                ins_pred = self.transform.label_encoder.decode(ins_pred)
                ins_pred_ents = get_entities_from_tag_seq(tokens, ins_pred)
                results.append(ins_pred_ents)
        return results
